# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import re

import requests

from ..common import NotifyImageSize, NotifyType
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from .base import NotifyBase


class FORMPayloadField:
    """Identifies the fields available in the FORM Payload."""

    VERSION = "version"
    TITLE = "title"
    MESSAGE = "message"
    MESSAGETYPE = "type"


# Defines the method to send the notification
METHODS = ("POST", "GET", "DELETE", "PUT", "HEAD", "PATCH")


class NotifyForm(NotifyBase):
    """A wrapper for Form Notifications."""

    # Support
    # - file*
    # - file?
    # - file*name
    # - file?name
    # - ?file
    # - *file
    # - file
    # The code will convert the ? or * to the digit increments
    __attach_as_re = re.compile(
        r"((?P<match1>(?P<id1a>[a-z0-9_-]+)?"
        r"(?P<wc1>[*?+$:.%]+)(?P<id1b>[a-z0-9_-]+))"
        r"|(?P<match2>(?P<id2>[a-z0-9_-]+)(?P<wc2>[*?+$:.%]?)))",
        re.IGNORECASE,
    )

    # Our count
    attach_as_count = "{:02d}"

    # the default attach_as value
    attach_as_default = f"file{attach_as_count}"

    # The default descriptive name associated with the Notification
    service_name = "Form"

    # The default protocol
    protocol = "form"

    # The default secure protocol
    secure_protocol = "forms"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_Custom_Form"

    # Support attachments
    attachment_support = True

    # Allows the user to specify the NotifyImageSize object
    image_size = NotifyImageSize.XY_128

    # Disable throttle rate for Form requests since they are normally
    # local anyway
    request_rate_per_sec = 0

    # Define the FORM version to place in all payloads
    # Version: Major.Minor,  Major is only updated if the entire schema is
    # changed. If just adding new items (or removing old ones, only increment
    # the Minor!
    form_version = "1.0"

    # Define object templates
    templates = (
        "{schema}://{host}",
        "{schema}://{host}:{port}",
        "{schema}://{user}@{host}",
        "{schema}://{user}@{host}:{port}",
        "{schema}://{user}:{password}@{host}",
        "{schema}://{user}:{password}@{host}:{port}",
    )

    # Define our tokens; these are the minimum tokens required required to
    # be passed into this function (as arguments). The syntax appends any
    # previously defined in the base package and builds onto them
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "host": {
                "name": _("Hostname"),
                "type": "string",
                "required": True,
            },
            "port": {
                "name": _("Port"),
                "type": "int",
                "min": 1,
                "max": 65535,
            },
            "user": {
                "name": _("Username"),
                "type": "string",
            },
            "password": {
                "name": _("Password"),
                "type": "string",
                "private": True,
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "method": {
                "name": _("Fetch Method"),
                "type": "choice:string",
                "values": METHODS,
                "default": METHODS[0],
            },
            "attach-as": {
                "name": _("Attach File As"),
                "type": "string",
                "default": "file*",
                "map_to": "attach_as",
            },
        },
    )

    # Define any kwargs we're using
    template_kwargs = {
        "headers": {
            "name": _("HTTP Header"),
            "prefix": "+",
        },
        "payload": {
            "name": _("Payload Extras"),
            "prefix": ":",
        },
        "params": {
            "name": _("GET Params"),
            "prefix": "-",
        },
    }

    def __init__(
        self,
        headers=None,
        method=None,
        payload=None,
        params=None,
        attach_as=None,
        **kwargs,
    ):
        """Initialize Form Object.

        headers can be a dictionary of key/value pairs that you want to
        additionally include as part of the server headers to post with
        """
        super().__init__(**kwargs)

        self.fullpath = kwargs.get("fullpath")
        if not isinstance(self.fullpath, str):
            self.fullpath = ""

        self.method = (
            self.template_args["method"]["default"]
            if not isinstance(method, str)
            else method.upper()
        )

        if self.method not in METHODS:
            msg = f"The method specified ({method}) is invalid."
            self.logger.warning(msg)
            raise TypeError(msg)

        # Custom File Attachment Over-Ride Support
        if not isinstance(attach_as, str):
            # Default value
            self.attach_as = self.attach_as_default
            self.attach_multi_support = True

        else:
            result = self.__attach_as_re.match(attach_as.strip())
            if not result:
                msg = f"The attach-as specified ({attach_as}) is invalid."
                self.logger.warning(msg)
                raise TypeError(msg)

            self.attach_as = ""
            self.attach_multi_support = False
            if result.group("match1"):
                if result.group("id1a"):
                    self.attach_as += result.group("id1a")

                self.attach_as += self.attach_as_count
                self.attach_multi_support = True
                self.attach_as += result.group("id1b")

            else:  # result.group('match2'):
                self.attach_as += result.group("id2")
                if result.group("wc2"):
                    self.attach_as += self.attach_as_count
                    self.attach_multi_support = True

        # A payload map allows users to over-ride the default mapping if
        # they're detected with the :overide=value.  Normally this would
        # create a new key and assign it the value specified.  However
        # if the key you specify is actually an internally mapped one,
        # then a re-mapping takes place using the value
        self.payload_map = {
            FORMPayloadField.VERSION: FORMPayloadField.VERSION,
            FORMPayloadField.TITLE: FORMPayloadField.TITLE,
            FORMPayloadField.MESSAGE: FORMPayloadField.MESSAGE,
            FORMPayloadField.MESSAGETYPE: FORMPayloadField.MESSAGETYPE,
        }

        self.params = {}
        if params:
            # Store our extra headers
            self.params.update(params)

        self.headers = {}
        if headers:
            # Store our extra headers
            self.headers.update(headers)

        self.payload_overrides = {}
        self.payload_extras = {}
        if payload:
            # Store our extra payload entries
            self.payload_extras.update(payload)
            for key in list(self.payload_extras.keys()):
                # Any values set in the payload to alter a system related one
                # alters the system key.  Hence :message=msg maps the 'message'
                # variable that otherwise already contains the payload to be
                # 'msg' instead (containing the payload)
                if key in self.payload_map:
                    self.payload_map[key] = self.payload_extras[key]
                    self.payload_overrides[key] = self.payload_extras[key]
                    del self.payload_extras[key]

        return

    def send(
        self,
        body,
        title="",
        notify_type=NotifyType.INFO,
        attach=None,
        **kwargs,
    ):
        """Perform Form Notification."""

        # Prepare HTTP Headers
        headers = {
            "User-Agent": self.app_id,
        }

        # Apply any/all header over-rides defined
        headers.update(self.headers)

        # Track our potential attachments
        files = []
        if attach and self.attachment_support:
            for no, attachment in enumerate(attach, start=1):
                # Perform some simple error checking
                if not attachment:
                    # We could not access the attachment
                    self.logger.error(
                        "Could not access attachment"
                        f" {attachment.url(privacy=True)}."
                    )
                    return False

                try:
                    files.append((
                        (
                            self.attach_as.format(no)
                            if self.attach_multi_support
                            else self.attach_as
                        ),
                        (
                            (
                                attachment.name
                                if attachment.name
                                else f"file{no:03}.dat"
                            ),
                            # file handle is safely closed in `finally`; inline
                            # open is intentional
                            open(attachment.path, "rb"),  # noqa: SIM115
                            attachment.mimetype,
                        ),
                    ))

                except OSError as e:
                    self.logger.warning(
                        "An I/O error occurred while opening {}.".format(
                            attachment.name if attachment else "attachment"
                        )
                    )
                    self.logger.debug(f"I/O Exception: {e!s}")
                    return False

            if not self.attach_multi_support and no > 1:
                self.logger.warning(
                    "Multiple attachments provided while "
                    "form:// Multi-Attachment Support not enabled"
                )

        # prepare Form Object
        payload = {}

        for key, value in (
            (FORMPayloadField.VERSION, self.form_version),
            (FORMPayloadField.TITLE, title),
            (FORMPayloadField.MESSAGE, body),
            (FORMPayloadField.MESSAGETYPE, notify_type),
        ):

            if not self.payload_map[key]:
                # Do not store element in payload response
                continue
            payload[self.payload_map[key]] = value

        # Apply any/all payload over-rides defined
        payload.update(self.payload_extras)

        auth = None
        if self.user:
            auth = (self.user, self.password)

        # Set our schema
        schema = "https" if self.secure else "http"

        url = f"{schema}://{self.host}"
        if isinstance(self.port, int):
            url += f":{self.port}"

        url += self.fullpath

        self.logger.debug(
            f"Form {self.method} URL:"
            f" {url} (cert_verify={self.verify_certificate!r})"
        )
        self.logger.debug(f"Form Payload: {payload!s}")

        # Always call throttle before any remote server i/o is made
        self.throttle()

        if self.method == "GET":
            method = requests.get
            payload.update(self.params)

        elif self.method == "PUT":
            method = requests.put

        elif self.method == "PATCH":
            method = requests.patch

        elif self.method == "DELETE":
            method = requests.delete

        elif self.method == "HEAD":
            method = requests.head

        else:  # POST
            method = requests.post

        try:
            r = method(
                url,
                files=files if files else None,
                data=payload if self.method != "GET" else None,
                params=payload if self.method == "GET" else self.params,
                headers=headers,
                auth=auth,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )
            if r.status_code < 200 or r.status_code >= 300:
                # We had a problem
                status_str = NotifyForm.http_response_code_lookup(
                    r.status_code
                )

                self.logger.warning(
                    "Failed to send Form %s notification: %s%serror=%s.",
                    self.method,
                    status_str,
                    ", " if status_str else "",
                    str(r.status_code),
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                # Return; we're done
                return False

            else:
                self.logger.info("Sent Form %s notification.", self.method)

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred sending Form "
                f"notification to {self.host}."
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            # Return; we're done
            return False

        except OSError as e:
            self.logger.warning(
                "An I/O error occurred while reading one of the "
                "attached files."
            )
            self.logger.debug(f"I/O Exception: {e!s}")
            return False

        finally:
            for file in files:
                # Ensure all files are closed
                file[1][1].close()

        return True

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (
            self.secure_protocol if self.secure else self.protocol,
            self.user,
            self.password,
            self.host,
            self.port if self.port else (443 if self.secure else 80),
            self.fullpath.rstrip("/"),
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "method": self.method,
        }

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        # Append our headers into our parameters
        params.update({f"+{k}": v for k, v in self.headers.items()})

        # Append our GET params into our parameters
        params.update({f"-{k}": v for k, v in self.params.items()})

        # Append our payload extra's into our parameters
        params.update({f":{k}": v for k, v in self.payload_extras.items()})
        params.update({f":{k}": v for k, v in self.payload_overrides.items()})

        if self.attach_as != self.attach_as_default:
            # Provide Attach-As extension details
            params["attach-as"] = self.attach_as

        # Determine Authentication
        auth = ""
        if self.user and self.password:
            auth = "{user}:{password}@".format(
                user=NotifyForm.quote(self.user, safe=""),
                password=self.pprint(
                    self.password, privacy, mode=PrivacyMode.Secret, safe=""
                ),
            )
        elif self.user:
            auth = "{user}@".format(
                user=NotifyForm.quote(self.user, safe=""),
            )

        default_port = 443 if self.secure else 80
        return "{schema}://{auth}{hostname}{port}{fullpath}?{params}".format(
            schema=self.secure_protocol if self.secure else self.protocol,
            auth=auth,
            # never encode hostname since we're expecting it to be a valid one
            hostname=self.host,
            port=(
                ""
                if self.port is None or self.port == default_port
                else f":{self.port}"
            ),
            fullpath=(
                NotifyForm.quote(self.fullpath, safe="/")
                if self.fullpath
                else "/"
            ),
            params=NotifyForm.urlencode(params),
        )

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # store any additional payload extra's defined
        results["payload"] = {
            NotifyForm.unquote(x): NotifyForm.unquote(y)
            for x, y in results["qsd:"].items()
        }

        # Add our headers that the user can potentially over-ride if they wish
        # to to our returned result set and tidy entries by unquoting them
        results["headers"] = {
            NotifyForm.unquote(x): NotifyForm.unquote(y)
            for x, y in results["qsd+"].items()
        }

        # Add our GET paramters in the event the user wants to pass these along
        results["params"] = {
            NotifyForm.unquote(x): NotifyForm.unquote(y)
            for x, y in results["qsd-"].items()
        }

        # Allow Attach-As Support which over-rides the name of the filename
        # posted with the form://
        # the default is file01, file02, file03, etc
        if "attach-as" in results["qsd"] and len(results["qsd"]["attach-as"]):
            results["attach_as"] = results["qsd"]["attach-as"]

        # Set method if not otherwise set
        if "method" in results["qsd"] and len(results["qsd"]["method"]):
            results["method"] = NotifyForm.unquote(results["qsd"]["method"])

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# For this to work correctly you need to register an app
# and generate an access token
#
#
#  This plugin will simply work using the url of:
#     streamlabs://access_token/
#
# API Documentation on Webhooks:
#    - https://dev.streamlabs.com/
#
import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import validate_regex
from .base import NotifyBase


# calls
class StrmlabsCall:
    ALERT = "ALERTS"
    DONATION = "DONATIONS"


# A List of calls we can use for verification
STRMLABS_CALLS = (
    StrmlabsCall.ALERT,
    StrmlabsCall.DONATION,
)


# alerts
class StrmlabsAlert:
    FOLLOW = "follow"
    SUBSCRIPTION = "subscription"
    DONATION = "donation"
    HOST = "host"


# A List of calls we can use for verification
STRMLABS_ALERTS = (
    StrmlabsAlert.FOLLOW,
    StrmlabsAlert.SUBSCRIPTION,
    StrmlabsAlert.DONATION,
    StrmlabsAlert.HOST,
)


class NotifyStreamlabs(NotifyBase):
    """A wrapper to Streamlabs Donation Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Streamlabs"

    # The services URL
    service_url = "https://streamlabs.com/"

    # The default secure protocol
    secure_protocol = "strmlabs"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_streamlabs"

    # Streamlabs Api endpoint
    notify_url = "https://streamlabs.com/api/v1.0/"

    # The maximum allowable characters allowed in the body per message
    body_maxlen = 255

    # Define object templates
    templates = ("{schema}://{access_token}/",)

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "access_token": {
                "name": _("Access Token"),
                "private": True,
                "required": True,
                "type": "string",
                "regex": (r"^[a-z0-9]{40}$", "i"),
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "call": {
                "name": _("Call"),
                "type": "choice:string",
                "values": STRMLABS_CALLS,
                "default": StrmlabsCall.ALERT,
            },
            "alert_type": {
                "name": _("Alert Type"),
                "type": "choice:string",
                "values": STRMLABS_ALERTS,
                "default": StrmlabsAlert.DONATION,
            },
            "image_href": {
                "name": _("Image Link"),
                "type": "string",
                "default": "",
            },
            "sound_href": {
                "name": _("Sound Link"),
                "type": "string",
                "default": "",
            },
            "duration": {
                "name": _("Duration"),
                "type": "int",
                "default": 1000,
                "min": 0,
            },
            "special_text_color": {
                "name": _("Special Text Color"),
                "type": "string",
                "default": "",
                "regex": (r"^[A-Z]$", "i"),
            },
            "amount": {
                "name": _("Amount"),
                "type": "int",
                "default": 0,
                "min": 0,
            },
            "currency": {
                "name": _("Currency"),
                "type": "string",
                "default": "USD",
                "regex": (r"^[A-Z]{3}$", "i"),
            },
            "name": {
                "name": _("Name"),
                "type": "string",
                "default": "Anon",
                "regex": (r"^[^\s].{1,24}$", "i"),
            },
            "identifier": {
                "name": _("Identifier"),
                "type": "string",
                "default": "Apprise",
            },
        },
    )

    def __init__(
        self,
        access_token,
        call=StrmlabsCall.ALERT,
        alert_type=StrmlabsAlert.DONATION,
        image_href="",
        sound_href="",
        duration=1000,
        special_text_color="",
        amount=0,
        currency="USD",
        name="Anon",
        identifier="Apprise",
        **kwargs,
    ):
        """Initialize Streamlabs Object."""
        super().__init__(**kwargs)

        # access token is generated by user
        # using https://streamlabs.com/api/v1.0/token
        # Tokens for Streamlabs never need to be refreshed.
        self.access_token = validate_regex(
            access_token, *self.template_tokens["access_token"]["regex"]
        )
        if not self.access_token:
            msg = "An invalid Streamslabs access token was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        # Store the call
        try:
            if call not in STRMLABS_CALLS:
                # allow the outer except to handle this common response
                raise
            else:
                self.call = call
        except Exception as e:
            # Invalid region specified
            msg = f"The streamlabs call specified ({call}) is invalid."
            self.logger.warning(msg)
            self.logger.debug(f"Socket Exception: {e!s}")
            raise TypeError(msg) from None

        # Store the alert_type
        # only applicable when calling /alerts
        try:
            if alert_type not in STRMLABS_ALERTS:
                # allow the outer except to handle this common response
                raise
            else:
                self.alert_type = alert_type
        except Exception as e:
            # Invalid region specified
            msg = f"The streamlabs alert type specified ({call}) is invalid."
            self.logger.warning(msg)
            self.logger.debug(f"Socket Exception: {e!s}")
            raise TypeError(msg) from None

        # params only applicable when calling /alerts
        self.image_href = image_href
        self.sound_href = sound_href
        self.duration = duration
        self.special_text_color = special_text_color

        # only applicable when calling /donations
        # The amount of this donation.
        self.amount = amount

        # only applicable when calling /donations
        # The 3 letter currency code for this donation.
        # Must be one of the supported currency codes.
        self.currency = validate_regex(
            currency, *self.template_args["currency"]["regex"]
        )

        # only applicable when calling /donations
        if not self.currency:
            msg = "An invalid Streamslabs currency was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        # only applicable when calling /donations
        # The name of the donor
        self.name = validate_regex(name, *self.template_args["name"]["regex"])
        if not self.name:
            msg = "An invalid Streamslabs donor was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        # An identifier for this donor,
        # which is used to group donations with the same donor.
        # only applicable when calling /donations
        self.identifier = identifier

        return

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform Streamlabs notification call (either donation or alert)"""

        headers = {
            "User-Agent": self.app_id,
        }
        if self.call == StrmlabsCall.ALERT:

            data = {
                "access_token": self.access_token,
                "type": self.alert_type.lower(),
                "image_href": self.image_href,
                "sound_href": self.sound_href,
                "message": title,
                "user_massage": body,
                "duration": self.duration,
                "special_text_color": self.special_text_color,
            }

            try:
                r = requests.post(
                    self.notify_url + self.call.lower(),
                    headers=headers,
                    data=data,
                    verify=self.verify_certificate,
                )
                if r.status_code != requests.codes.ok:
                    # We had a problem
                    status_str = NotifyStreamlabs.http_response_code_lookup(
                        r.status_code
                    )

                    self.logger.warning(
                        "Failed to send Streamlabs alert: "
                        "{}{}error={}.".format(
                            status_str,
                            ", " if status_str else "",
                            r.status_code,
                        )
                    )

                    self.logger.debug(f"Response Details:\r\n{r.content}")
                    return False

                else:
                    self.logger.info("Sent Streamlabs alert.")

            except requests.RequestException as e:
                self.logger.warning(
                    "A Connection error occurred sending Streamlabs alert."
                )
                self.logger.debug(f"Socket Exception: {e!s}")
                return False

        if self.call == StrmlabsCall.DONATION:
            data = {
                "name": self.name,
                "identifier": self.identifier,
                "amount": self.amount,
                "currency": self.currency,
                "access_token": self.access_token,
                "message": body,
            }

            try:
                r = requests.post(
                    self.notify_url + self.call.lower(),
                    headers=headers,
                    data=data,
                    verify=self.verify_certificate,
                )
                if r.status_code != requests.codes.ok:
                    # We had a problem
                    status_str = NotifyStreamlabs.http_response_code_lookup(
                        r.status_code
                    )

                    self.logger.warning(
                        "Failed to send Streamlabs donation: "
                        "{}{}error={}.".format(
                            status_str,
                            ", " if status_str else "",
                            r.status_code,
                        )
                    )

                    self.logger.debug(f"Response Details:\r\n{r.content}")
                    return False

                else:
                    self.logger.info("Sent Streamlabs donation.")

            except requests.RequestException as e:
                self.logger.warning(
                    "A Connection error occurred sending Streamlabs donation."
                )
                self.logger.debug(f"Socket Exception: {e!s}")
                return False

        return True

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (self.secure_protocol, self.access_token)

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "call": self.call,
            # donation
            "name": self.name,
            "identifier": self.identifier,
            "amount": self.amount,
            "currency": self.currency,
            # alert
            "alert_type": self.alert_type,
            "image_href": self.image_href,
            "sound_href": self.sound_href,
            "duration": self.duration,
            "special_text_color": self.special_text_color,
        }

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))
        return "{schema}://{access_token}/?{params}".format(
            schema=self.secure_protocol,
            access_token=self.pprint(self.access_token, privacy, safe=""),
            params=NotifyStreamlabs.urlencode(params),
        )

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object.

        Syntax:
          strmlabs://access_token
        """
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Store our access code
        access_token = NotifyStreamlabs.unquote(results["host"])
        results["access_token"] = access_token

        # call
        if "call" in results["qsd"] and results["qsd"]["call"]:
            results["call"] = NotifyStreamlabs.unquote(
                results["qsd"]["call"].strip().upper()
            )
        # donation - amount
        if "amount" in results["qsd"] and results["qsd"]["amount"]:
            results["amount"] = NotifyStreamlabs.unquote(
                results["qsd"]["amount"]
            )
        # donation - currency
        if "currency" in results["qsd"] and results["qsd"]["currency"]:
            results["currency"] = NotifyStreamlabs.unquote(
                results["qsd"]["currency"].strip().upper()
            )
        # donation - name
        if "name" in results["qsd"] and results["qsd"]["name"]:
            results["name"] = NotifyStreamlabs.unquote(
                results["qsd"]["name"].strip().upper()
            )
        # donation - identifier
        if "identifier" in results["qsd"] and results["qsd"]["identifier"]:
            results["identifier"] = NotifyStreamlabs.unquote(
                results["qsd"]["identifier"].strip().upper()
            )
        # alert - alert_type
        if "alert_type" in results["qsd"] and results["qsd"]["alert_type"]:
            results["alert_type"] = NotifyStreamlabs.unquote(
                results["qsd"]["alert_type"]
            )
        # alert - image_href
        if "image_href" in results["qsd"] and results["qsd"]["image_href"]:
            results["image_href"] = NotifyStreamlabs.unquote(
                results["qsd"]["image_href"]
            )
        # alert - sound_href
        if "sound_href" in results["qsd"] and results["qsd"]["sound_href"]:
            results["sound_href"] = NotifyStreamlabs.unquote(
                results["qsd"]["sound_href"].strip().upper()
            )
        # alert - duration
        if "duration" in results["qsd"] and results["qsd"]["duration"]:
            results["duration"] = NotifyStreamlabs.unquote(
                results["qsd"]["duration"].strip().upper()
            )
        # alert - special_text_color
        if (
            "special_text_color" in results["qsd"]
            and results["qsd"]["special_text_color"]
        ):
            results["special_text_color"] = NotifyStreamlabs.unquote(
                results["qsd"]["special_text_color"].strip().upper()
            )

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import re

import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import validate_regex
from .base import NotifyBase

# Register at https://sct.ftqq.com/
#   - do as the page describe and you will get the token

# Syntax:
#  schan://{access_token}/


class NotifyServerChan(NotifyBase):
    """A wrapper for ServerChan Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "ServerChan"

    # The services URL
    service_url = "https://sct.ftqq.com/"

    # All notification requests are secure
    secure_protocol = "schan"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_serverchan"

    # ServerChan API
    notify_url = "https://sctapi.ftqq.com/{token}.send"

    # Define object templates
    templates = ("{schema}://{token}",)

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "token": {
                "name": _("Token"),
                "type": "string",
                "private": True,
                "required": True,
                "regex": (r"^[a-z0-9-]+$", "i"),
            },
        },
    )

    def __init__(self, token, **kwargs):
        """Initialize ServerChan Object."""
        super().__init__(**kwargs)

        # Token (associated with project)
        self.token = validate_regex(
            token, *self.template_tokens["token"]["regex"]
        )
        if not self.token:
            msg = f"An invalid ServerChan API Token ({token}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform ServerChan Notification."""
        payload = {
            "title": title,
            "desp": body,
        }

        # Our Notification URL
        notify_url = self.notify_url.format(token=self.token)

        # Some Debug Logging
        self.logger.debug(
            "ServerChan URL:"
            f" {notify_url} (cert_verify={self.verify_certificate})"
        )
        self.logger.debug(f"ServerChan Payload: {payload}")

        # Always call throttle before any remote server i/o is made
        self.throttle()

        try:
            r = requests.post(
                notify_url,
                data=payload,
            )

            if r.status_code != requests.codes.ok:
                # We had a problem
                status_str = NotifyServerChan.http_response_code_lookup(
                    r.status_code
                )

                self.logger.warning(
                    "Failed to send ServerChan notification: "
                    "{}{}error={}.".format(
                        status_str, ", " if status_str else "", r.status_code
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")
                return False

            else:
                self.logger.info("Sent ServerChan notification.")

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occured sending ServerChan notification."
            )
            self.logger.debug(f"Socket Exception: {e!s}")
            return False

        return True

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (self.secure_protocol, self.token)

    def url(self, privacy=False):
        """Returns the URL built dynamically based on specified arguments."""

        return "{schema}://{token}".format(
            schema=self.secure_protocol,
            token=self.pprint(self.token, privacy, safe=""),
        )

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to
        substantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't parse the URL
            return results

        pattern = "schan://([a-zA-Z0-9]+)/" + (
            "?" if not url.endswith("/") else ""
        )
        result = re.match(pattern, url)
        results["token"] = result.group(1) if result else ""
        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
import base64
import binascii
import copy
import json


def base64_urlencode(data: bytes) -> str:
    """URL Safe Base64 Encoding."""
    try:
        return base64.urlsafe_b64encode(data).rstrip(b"=").decode("utf-8")

    except TypeError:
        # data is not supported; avoid raising exception
        return None


def base64_urldecode(data: str) -> bytes:
    """URL Safe Base64 Encoding."""

    try:
        # Normalize base64url string (remove padding, add it back)
        padding = "=" * (-len(data) % 4)
        return base64.urlsafe_b64decode(data + padding)

    except TypeError:
        # data is not supported; avoid raising exception
        return None


def decode_b64_dict(di: dict) -> dict:
    """Decodes base64 dictionary previously encoded.

    string entries prefixed with `b64:` are targeted
    """
    di = copy.deepcopy(di)
    for k, v in di.items():
        if not isinstance(v, str) or not v.startswith("b64:"):
            continue

        try:
            parsed_v = base64.b64decode(v[4:])
            parsed_v = json.loads(parsed_v)

        except (
            ValueError,
            TypeError,
            binascii.Error,
            json.decoder.JSONDecodeError,
        ):
            # ValueError: the length of altchars is not 2.
            # TypeError: invalid input
            # binascii.Error: not base64 (bad padding)
            # json.decoder.JSONDecodeError: Bad JSON object

            parsed_v = v
        di[k] = parsed_v
    return di


def encode_b64_dict(di: dict, encoding="utf-8") -> tuple[dict, bool]:
    """Encodes dictionary entries containing binary types (int, float) into
    base64.

    Final product is always string based values
    """
    di = copy.deepcopy(di)
    needs_decoding = False
    for k, v in di.items():
        if isinstance(v, str):
            continue

        try:
            encoded = base64.urlsafe_b64encode(json.dumps(v).encode(encoding))
            encoded = f"b64:{encoded.decode(encoding)}"
            needs_decoding = True

        except (ValueError, TypeError):
            # ValueError:
            #  - the length of altchars is not 2.
            # TypeError:
            #  - json not searializable or
            #  - bytes object not passed into urlsafe_b64encode()
            encoded = str(v)

        di[k] = encoded
    return di, needs_decoding



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# To use this service you will need a D7 Networks account from their website
# at https://d7networks.com/
#
# After you've established your account you can get your api login credentials
# (both user and password) from the API Details section from within your
# account profile area:  https://d7networks.com/accounts/profile/
#
# API Reference: https://d7networks.com/docs/Messages/Send_Message/

from json import dumps, loads

import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import (
    is_phone_no,
    parse_bool,
    parse_phone_no,
    validate_regex,
)
from .base import NotifyBase

# Extend HTTP Error Messages
D7NETWORKS_HTTP_ERROR_MAP = {
    401: "Invalid Argument(s) Specified.",
    403: "Unauthorized - Authentication Failure.",
    412: "A Routing Error Occured",
    500: "A Serverside Error Occured Handling the Request.",
}


class NotifyD7Networks(NotifyBase):
    """A wrapper for D7 Networks Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "D7 Networks"

    # The services URL
    service_url = "https://d7networks.com/"

    # All notification requests are secure
    secure_protocol = "d7sms"

    # Allow 300 requests per minute.
    # 60/300 = 0.2
    request_rate_per_sec = 0.20

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_d7networks"

    # D7 Networks single notification URL
    notify_url = "https://api.d7networks.com/messages/v1/send"

    # The maximum length of the body
    body_maxlen = 160

    # A title can not be used for SMS Messages.  Setting this to zero will
    # cause any title (if defined) to get placed into the message body.
    title_maxlen = 0

    # Define object templates
    templates = ("{schema}://{token}@{targets}",)

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "token": {
                "name": _("API Access Token"),
                "type": "string",
                "required": True,
                "private": True,
            },
            "target_phone": {
                "name": _("Target Phone No"),
                "type": "string",
                "prefix": "+",
                "regex": (r"^[0-9\s)(+-]+$", "i"),
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
                "required": True,
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "unicode": {
                # Unicode characters (default is 'auto')
                "name": _("Unicode Characters"),
                "type": "bool",
                "default": False,
            },
            "batch": {
                "name": _("Batch Mode"),
                "type": "bool",
                "default": False,
            },
            "to": {
                "alias_of": "targets",
            },
            "source": {
                # Originating address,In cases where the rewriting of the
                # sender's address is supported or permitted by the SMS-C.
                # This is used to transmit the message, this number is
                # transmitted as the originating address and is completely
                # optional.
                "name": _("Originating Address"),
                "type": "string",
                "map_to": "source",
            },
            "from": {
                "alias_of": "source",
            },
        },
    )

    def __init__(
        self,
        token=None,
        targets=None,
        source=None,
        batch=False,
        unicode=None,
        **kwargs,
    ):
        """Initialize D7 Networks Object."""
        super().__init__(**kwargs)

        # Prepare Batch Mode Flag
        self.batch = batch

        # Setup our source address (if defined)
        self.source = None if not isinstance(source, str) else source.strip()

        # Define whether or not we should set the unicode flag
        self.unicode = (
            self.template_args["unicode"]["default"]
            if unicode is None
            else bool(unicode)
        )

        # The token associated with the account
        self.token = validate_regex(token)
        if not self.token:
            msg = f"The D7 Networks token specified ({token}) is invalid."
            self.logger.warning(msg)
            raise TypeError(msg)

        # Parse our targets
        self.targets = []
        for target in parse_phone_no(targets):
            # Validate targets and drop bad ones:
            result = result = is_phone_no(target)
            if not result:
                self.logger.warning(
                    f"Dropped invalid phone # ({target}) specified.",
                )
                continue

            # store valid phone number
            self.targets.append(result["full"])

        return

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Depending on whether we are set to batch mode or single mode this
        redirects to the appropriate handling."""

        if len(self.targets) == 0:
            # There were no services to notify
            self.logger.warning("There were no D7 Networks targets to notify.")
            return False

        # error tracking (used for function return)
        has_error = False

        # Prepare our headers
        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": f"Bearer {self.token}",
        }

        payload = {
            "message_globals": {
                "channel": "sms",
            },
            "messages": [{
                # Populated later on
                "recipients": None,
                "content": body,
                "data_coding":
                # auto is a better substitute over 'text' as text is easier to
                # detect from a post than `unicode` is.
                "auto" if not self.unicode else "unicode",
            }],
        }

        # use the list directly
        targets = list(self.targets)

        if self.source:
            payload["message_globals"]["originator"] = self.source

        target = None
        while len(targets):

            if self.batch:
                # Prepare our payload
                payload["messages"][0]["recipients"] = self.targets

                # Reset our targets so we don't keep going. This is required
                # because we're in batch mode; we only need to loop once.
                targets = []

            else:
                # We're not in a batch mode; so get our next target
                # Get our target(s) to notify
                target = targets.pop(0)

                # Prepare our payload
                payload["messages"][0]["recipients"] = [target]

            # Some Debug Logging
            self.logger.debug(
                "D7 Networks POST URL:"
                f" {self.notify_url} (cert_verify={self.verify_certificate})"
            )
            self.logger.debug(f"D7 Networks Payload: {payload}")

            # Always call throttle before any remote server i/o is made
            self.throttle()
            try:
                r = requests.post(
                    self.notify_url,
                    data=dumps(payload),
                    headers=headers,
                    verify=self.verify_certificate,
                    timeout=self.request_timeout,
                )

                if r.status_code not in (
                    requests.codes.created,
                    requests.codes.ok,
                ):
                    # We had a problem
                    status_str = NotifyBase.http_response_code_lookup(
                        r.status_code, D7NETWORKS_HTTP_ERROR_MAP
                    )

                    try:
                        # Update our status response if we can
                        json_response = loads(r.content)
                        status_str = json_response.get("message", status_str)

                    except (AttributeError, TypeError, ValueError):
                        # ValueError = r.content is Unparsable
                        # TypeError = r.content is None
                        # AttributeError = r is None

                        # We could not parse JSON response.
                        # We will just use the status we already have.
                        pass

                    self.logger.warning(
                        "Failed to send D7 Networks SMS notification to {}: "
                        "{}{}error={}.".format(
                            ", ".join(target) if self.batch else target,
                            status_str,
                            ", " if status_str else "",
                            r.status_code,
                        )
                    )

                    self.logger.debug(f"Response Details:\r\n{r.content}")

                    # Mark our failure
                    has_error = True
                    continue

                else:

                    if self.batch:
                        self.logger.info(
                            "Sent D7 Networks batch SMS notification to "
                            f"{len(self.targets)} target(s)."
                        )

                    else:
                        self.logger.info(
                            f"Sent D7 Networks SMS notification to {target}."
                        )

                    self.logger.debug(f"Response Details:\r\n{r.content}")

            except requests.RequestException as e:
                self.logger.warning(
                    "A Connection error occurred sending D7 Networks:{} "
                    .format(", ".join(self.targets))
                    + "notification."
                )
                self.logger.debug(f"Socket Exception: {e!s}")
                # Mark our failure
                has_error = True
                continue

        return not has_error

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "batch": "yes" if self.batch else "no",
            "unicode": "yes" if self.unicode else "no",
        }

        if self.source:
            params["from"] = self.source

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        return "{schema}://{token}@{targets}/?{params}".format(
            schema=self.secure_protocol,
            token=self.pprint(self.token, privacy, safe=""),
            targets="/".join(
                [NotifyD7Networks.quote(x, safe="") for x in self.targets]
            ),
            params=NotifyD7Networks.urlencode(params),
        )

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (self.secure_protocol, self.token)

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        #
        # Factor batch into calculation
        #
        return len(self.targets) if not self.batch else 1

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        if "token" in results["qsd"] and len(results["qsd"]["token"]):
            results["token"] = NotifyD7Networks.unquote(
                results["qsd"]["token"]
            )

        elif results["user"]:
            results["token"] = NotifyD7Networks.unquote(results["user"])

            if results["password"]:
                # Support token containing a colon (:)
                results["token"] += ":" + NotifyD7Networks.unquote(
                    results["password"]
                )

        elif results["password"]:
            # Support token starting with a colon (:)
            results["token"] = ":" + NotifyD7Networks.unquote(
                results["password"]
            )

        # Initialize our targets
        results["targets"] = []

        # The store our first target stored in the hostname
        results["targets"].append(NotifyD7Networks.unquote(results["host"]))

        # Get our entries; split_path() looks after unquoting content for us
        # by default
        results["targets"].extend(
            NotifyD7Networks.split_path(results["fullpath"])
        )

        # Get Batch Mode Flag
        results["batch"] = parse_bool(results["qsd"].get("batch", False))

        # Get Unicode Flag
        results["unicode"] = parse_bool(results["qsd"].get("unicode", False))

        # Support the 'to' variable so that we can support targets this way too
        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += NotifyD7Networks.parse_phone_no(
                results["qsd"]["to"]
            )

        # Support the 'from' and source variable
        if "from" in results["qsd"] and len(results["qsd"]["from"]):
            results["source"] = NotifyD7Networks.unquote(
                results["qsd"]["from"]
            )

        elif "source" in results["qsd"] and len(results["qsd"]["source"]):
            results["source"] = NotifyD7Networks.unquote(
                results["qsd"]["source"]
            )

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from json import dumps

import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from ..utils.parse import validate_regex
from .base import NotifyBase


class NotifyPushjet(NotifyBase):
    """A wrapper for Pushjet Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Pushjet"

    # The default protocol
    protocol = "pjet"

    # The default secure protocol
    secure_protocol = "pjets"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_pushjet"

    # Disable throttle rate for Pushjet requests since they are normally
    # local anyway (the remote/online service is no more)
    request_rate_per_sec = 0

    # Define object templates
    templates = (
        "{schema}://{host}:{port}/{secret_key}",
        "{schema}://{host}/{secret_key}",
        "{schema}://{user}:{password}@{host}:{port}/{secret_key}",
        "{schema}://{user}:{password}@{host}/{secret_key}",
    )

    # Define our tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "host": {
                "name": _("Hostname"),
                "type": "string",
                "required": True,
            },
            "port": {
                "name": _("Port"),
                "type": "int",
                "min": 1,
                "max": 65535,
            },
            "secret_key": {
                "name": _("Secret Key"),
                "type": "string",
                "required": True,
                "private": True,
            },
            "user": {
                "name": _("Username"),
                "type": "string",
            },
            "password": {
                "name": _("Password"),
                "type": "string",
                "private": True,
            },
        },
    )

    template_args = dict(
        NotifyBase.template_args,
        **{
            "secret": {
                "alias_of": "secret_key",
            },
        },
    )

    def __init__(self, secret_key, **kwargs):
        """Initialize Pushjet Object."""
        super().__init__(**kwargs)

        # Secret Key (associated with project)
        self.secret_key = validate_regex(secret_key)
        if not self.secret_key:
            msg = (
                f"An invalid Pushjet Secret Key ({secret_key}) was specified."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        return

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (
            self.secure_protocol if self.secure else self.protocol,
            self.user,
            self.password,
            self.host,
            self.port,
            self.secret_key,
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Our URL parameters
        params = self.url_parameters(privacy=privacy, *args, **kwargs)

        default_port = 443 if self.secure else 80

        # Determine Authentication
        auth = ""
        if self.user and self.password:
            auth = "{user}:{password}@".format(
                user=NotifyPushjet.quote(self.user, safe=""),
                password=self.pprint(
                    self.password, privacy, mode=PrivacyMode.Secret, safe=""
                ),
            )

        return "{schema}://{auth}{hostname}{port}/{secret}/?{params}".format(
            schema=self.secure_protocol if self.secure else self.protocol,
            auth=auth,
            # never encode hostname since we're expecting it to be a valid one
            hostname=self.host,
            port=(
                ""
                if self.port is None or self.port == default_port
                else f":{self.port}"
            ),
            secret=self.pprint(
                self.secret_key, privacy, mode=PrivacyMode.Secret, safe=""
            ),
            params=NotifyPushjet.urlencode(params),
        )

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform Pushjet Notification."""

        params = {
            "secret": self.secret_key,
        }

        # prepare Pushjet Object
        payload = {
            "message": body,
            "title": title,
            "link": None,
            "level": None,
        }

        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/x-www-form-urlencoded; charset=utf-8",
        }

        auth = None
        if self.user:
            auth = (self.user, self.password)

        notify_url = "{schema}://{host}{port}/message/".format(
            schema="https" if self.secure else "http",
            host=self.host,
            port=f":{self.port}" if self.port else "",
        )

        self.logger.debug(
            "Pushjet POST URL:"
            f" {notify_url} (cert_verify={self.verify_certificate!r})"
        )
        self.logger.debug(f"Pushjet Payload: {payload!s}")

        # Always call throttle before any remote server i/o is made
        self.throttle()

        try:
            r = requests.post(
                notify_url,
                params=params,
                data=dumps(payload),
                headers=headers,
                auth=auth,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )
            if r.status_code != requests.codes.ok:
                # We had a problem
                status_str = NotifyPushjet.http_response_code_lookup(
                    r.status_code
                )

                self.logger.warning(
                    "Failed to send Pushjet notification: "
                    "{}{}error={}.".format(
                        status_str, ", " if status_str else "", r.status_code
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                # Return; we're done
                return False

            else:
                self.logger.info("Sent Pushjet notification.")

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred sending Pushjet "
                f"notification to {self.host}."
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            # Return; we're done
            return False

        return True

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object.

        Syntax:
           pjet://hostname/secret_key
           pjet://hostname:port/secret_key
           pjet://user:pass@hostname/secret_key
           pjet://user:pass@hostname:port/secret_key
           pjets://hostname/secret_key
           pjets://hostname:port/secret_key
           pjets://user:pass@hostname/secret_key
           pjets://user:pass@hostname:port/secret_key
        """
        results = NotifyBase.parse_url(url)
        if not results:
            # We're done early as we couldn't load the results
            return results

        try:
            # Retrieve our secret_key from the first entry in the url path
            results["secret_key"] = NotifyPushjet.split_path(
                results["fullpath"]
            )[0]

        except IndexError:
            # no secret key specified
            results["secret_key"] = None

        # Allow over-riding the secret by specifying it as an argument
        # this allows people who have http-auth infront to login
        # through it in addition to supporting the secret key
        if "secret" in results["qsd"] and len(results["qsd"]["secret"]):
            results["secret_key"] = NotifyPushjet.unquote(
                results["qsd"]["secret"]
            )

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# For this to work correctly you need to create a webhook. You'll also
# need a GSuite account (there are free trials if you don't have one)
#
#  - Open Google Chat in your browser:
#     Link: https://chat.google.com/
#  - Go to the room to which you want to add a bot.
#  - From the room menu at the top of the page, select Manage webhooks.
#  - Provide it a name and optional avatar and click SAVE
#  - Copy the URL listed next to your new webhook in the Webhook URL column.
#  - Click outside the dialog box to close.
#
# When you've completed, you'll get a URL that looks a little like this:
#  https://chat.googleapis.com/v1/spaces/AAAAk6lGXyM/\
#       messages?key=AIzaSyDdI0hCZtE6vySjMm-WEfRq3CPzqKqqsHI&\
#       token=O7b1nyri_waOpLMSzbFILAGRzgtQofPW71fEEXKcyFk%3D
#
# Simplified, it looks like this:
#     https://chat.googleapis.com/v1/spaces/WORKSPACE/messages?\
#       key=WEBHOOK_KEY&token=WEBHOOK_TOKEN
#
# This plugin will simply work using the url of:
#     gchat://WORKSPACE/WEBHOOK_KEY/WEBHOOK_TOKEN
#
# API Documentation on Webhooks:
#    - https://developers.google.com/hangouts/chat/quickstart/\
#         incoming-bot-python
#    - https://developers.google.com/hangouts/chat/reference/rest
#
from json import dumps
import re

import requests

from ..common import NotifyFormat, NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import validate_regex
from .base import NotifyBase


class NotifyGoogleChat(NotifyBase):
    """A wrapper to Google Chat Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Google Chat"

    # The services URL
    service_url = "https://chat.google.com/"

    # The default secure protocol
    secure_protocol = "gchat"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_googlechat"

    # Google Chat Webhook
    notify_url = "https://chat.googleapis.com/v1/spaces/{workspace}/messages"

    # Default Notify Format
    notify_format = NotifyFormat.MARKDOWN

    # A title can not be used for Google Chat Messages.  Setting this to zero
    # will cause any title (if defined) to get placed into the message body.
    title_maxlen = 0

    # The maximum allowable characters allowed in the body per message
    body_maxlen = 4000

    # Define object templates
    templates = (
        "{schema}://{workspace}/{webhook_key}/{webhook_token}",
        "{schema}://{workspace}/{webhook_key}/{webhook_token}/{thread_key}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "workspace": {
                "name": _("Workspace"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "webhook_key": {
                "name": _("Webhook Key"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "webhook_token": {
                "name": _("Webhook Token"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "thread_key": {
                "name": _("Thread Key"),
                "type": "string",
                "private": True,
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "workspace": {
                "alias_of": "workspace",
            },
            "key": {
                "alias_of": "webhook_key",
            },
            "token": {
                "alias_of": "webhook_token",
            },
            "thread": {
                "alias_of": "thread_key",
            },
        },
    )

    def __init__(
        self, workspace, webhook_key, webhook_token, thread_key=None, **kwargs
    ):
        """Initialize Google Chat Object."""
        super().__init__(**kwargs)

        # Workspace (associated with project)
        self.workspace = validate_regex(workspace)
        if not self.workspace:
            msg = (
                "An invalid Google Chat Workspace "
                f"({workspace}) was specified."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # Webhook Key (associated with project)
        self.webhook_key = validate_regex(webhook_key)
        if not self.webhook_key:
            msg = (
                "An invalid Google Chat Webhook Key "
                f"({webhook_key}) was specified."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # Webhook Token (associated with project)
        self.webhook_token = validate_regex(webhook_token)
        if not self.webhook_token:
            msg = (
                "An invalid Google Chat Webhook Token "
                f"({webhook_token}) was specified."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        if thread_key:
            self.thread_key = validate_regex(thread_key)
            if not self.thread_key:
                msg = (
                    "An invalid Google Chat Thread Key "
                    f"({thread_key}) was specified."
                )
                self.logger.warning(msg)
                raise TypeError(msg)
        else:
            self.thread_key = None

        return

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform Google Chat Notification."""

        # Our headers
        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json; charset=utf-8",
        }

        payload = {
            # Our Message
            "text": body,
        }

        # Construct Notify URL
        notify_url = self.notify_url.format(
            workspace=self.workspace,
        )

        params = {
            # Prepare our URL Parameters
            "token": self.webhook_token,
            "key": self.webhook_key,
        }

        if self.thread_key:
            params.update({
                "messageReplyOption": "REPLY_MESSAGE_FALLBACK_TO_NEW_THREAD",
            })

            payload.update({
                "thread": {
                    "thread_key": self.thread_key,
                }
            })

        self.logger.debug(
            "Google Chat POST URL:"
            f" {notify_url} (cert_verify={self.verify_certificate!r})"
        )
        self.logger.debug(f"Google Chat Parameters: {params!s}")
        self.logger.debug(f"Google Chat Payload: {payload!s}")

        # Always call throttle before any remote server i/o is made
        self.throttle()
        try:
            r = requests.post(
                notify_url,
                params=params,
                data=dumps(payload),
                headers=headers,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )
            if r.status_code not in (
                requests.codes.ok,
                requests.codes.no_content,
            ):

                # We had a problem
                status_str = NotifyBase.http_response_code_lookup(
                    r.status_code
                )

                self.logger.warning(
                    "Failed to send Google Chat notification: "
                    "{}{}error={}.".format(
                        status_str, ", " if status_str else "", r.status_code
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                # Return; we're done
                return False

            else:
                self.logger.info("Sent Google Chat notification.")

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred postingto Google Chat."
            )
            self.logger.debug(f"Socket Exception: {e!s}")
            return False

        return True

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (
            self.secure_protocol,
            self.workspace,
            self.webhook_key,
            self.webhook_token,
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Set our parameters
        params = self.url_parameters(privacy=privacy, *args, **kwargs)

        return "{schema}://{workspace}/{key}/{token}/{thread}?{params}".format(
            schema=self.secure_protocol,
            workspace=self.pprint(self.workspace, privacy, safe=""),
            key=self.pprint(self.webhook_key, privacy, safe=""),
            token=self.pprint(self.webhook_token, privacy, safe=""),
            thread=(
                ""
                if not self.thread_key
                else self.pprint(self.thread_key, privacy, safe="")
            ),
            params=NotifyGoogleChat.urlencode(params),
        )

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object.

        Syntax:
          gchat://workspace/webhook_key/webhook_token
          gchat://workspace/webhook_key/webhook_token/thread_key
        """
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Store our Workspace
        results["workspace"] = NotifyGoogleChat.unquote(results["host"])

        # Acquire our tokens
        tokens = NotifyGoogleChat.split_path(results["fullpath"])

        # Store our Webhook Key
        results["webhook_key"] = tokens.pop(0) if tokens else None

        # Store our Webhook Token
        results["webhook_token"] = tokens.pop(0) if tokens else None

        # Store our Thread Key
        results["thread_key"] = tokens.pop(0) if tokens else None

        # Support arguments as overrides (if specified)
        if "workspace" in results["qsd"]:
            results["workspace"] = NotifyGoogleChat.unquote(
                results["qsd"]["workspace"]
            )

        if "key" in results["qsd"]:
            results["webhook_key"] = NotifyGoogleChat.unquote(
                results["qsd"]["key"]
            )

        if "token" in results["qsd"]:
            results["webhook_token"] = NotifyGoogleChat.unquote(
                results["qsd"]["token"]
            )

        if "thread" in results["qsd"]:
            results["thread_key"] = NotifyGoogleChat.unquote(
                results["qsd"]["thread"]
            )

        elif "threadkey" in results["qsd"]:
            # Support Google Chat's Thread Key (if set)
            # keys are always made lowercase; so check above is attually
            # testing threadKey successfully as well
            results["thread_key"] = NotifyGoogleChat.unquote(
                results["qsd"]["threadkey"]
            )

        return results

    @staticmethod
    def parse_native_url(url):
        """
        Support
           https://chat.googleapis.com/v1/spaces/{workspace}/messages
                 '?key={key}&token={token}
           https://chat.googleapis.com/v1/spaces/{workspace}/messages
                 '?key={key}&token={token}&threadKey={thread}
        """

        result = re.match(
            r"^https://chat\.googleapis\.com/v1/spaces/"
            r"(?P<workspace>[A-Z0-9_-]+)/messages/*(?P<params>.+)$",
            url,
            re.I,
        )

        if result:
            return NotifyGoogleChat.parse_url(
                "{schema}://{workspace}/{params}".format(
                    schema=NotifyGoogleChat.secure_protocol,
                    workspace=result.group("workspace"),
                    params=result.group("params"),
                )
            )

        return None



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Simple API Reference:
#  - https://www.notificationapi.com/docs/reference/server#send

from __future__ import annotations

import base64
from email.utils import formataddr
from itertools import chain
from json import dumps, loads
import re

import requests

from ..common import NotifyFormat, NotifyImageSize, NotifyType
from ..conversion import convert_between
from ..locale import gettext_lazy as _
from ..utils.parse import (
    is_email,
    is_phone_no,
    parse_emails,
    parse_list,
    validate_regex,
)
from .base import NotifyBase

# Used to detect ID
IS_VALID_ID_RE = re.compile(
    r"^\s*(@|%40)?(?P<id>[\w_-]+)\s*$", re.I)


class NotificationAPIRegion:
    """Regions."""

    CA = "ca"
    US = "us"
    EU = "eu"


# NotificationAPI endpoints
NOTIFICATIONAPI_API_LOOKUP = {
    NotificationAPIRegion.US: "https://api.notificationapi.com",
    NotificationAPIRegion.CA: "https://api.ca.notificationapi.com",
    NotificationAPIRegion.EU: "https://api.eu.notificationapi.com",
}

# A List of our regions we can use for verification
NOTIFICATIONAPI_REGIONS = (
    NotificationAPIRegion.US,
    NotificationAPIRegion.CA,
    NotificationAPIRegion.EU,
)


class NotificationAPIChannel:
    """Channels"""

    EMAIL = "email"
    SMS = "sms"
    INAPP = "inapp"
    WEB_PUSH = "web_push"
    MOBILE_PUSH = "mobile_push"
    SLACK = "slack"


# A List of our channels we can use for verification
NOTIFICATIONAPI_CHANNELS: frozenset[str] = frozenset([
    NotificationAPIChannel.EMAIL,
    NotificationAPIChannel.SMS,
    NotificationAPIChannel.INAPP,
    NotificationAPIChannel.WEB_PUSH,
    NotificationAPIChannel.MOBILE_PUSH,
    NotificationAPIChannel.SLACK,
])


class NotificationAPIMode:
    """Modes"""

    TEMPLATE = "template"
    MESSAGE = "message"


# A List of our channels we can use for verification
NOTIFICATIONAPI_MODES: frozenset[str] = frozenset([
    NotificationAPIMode.TEMPLATE,
    NotificationAPIMode.MESSAGE,
])


class NotifyNotificationAPI(NotifyBase):
    """
    A wrapper for NotificationAPI Notifications
    """

    # The default descriptive name associated with the Notification
    service_name = "NotificationAPI"

    # The services URL
    service_url = "https://www.notificationapi.com/"

    # The default secure protocol
    secure_protocol = ("napi", "notificationapi")

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_notificationapi"

    # If no NotificationAPI Message Type is specified, then the following is
    # used
    default_message_type = "apprise"

    # Allow 300 requests per minute.
    # 60/300 = 0.2
    request_rate_per_sec = 0.2

    # Allows the user to specify the NotifyImageSize object
    image_size = NotifyImageSize.XY_72

    # Define object templates
    templates = (
        "{schema}://{client_id}/{client_secret}/{targets}",
        "{schema}://{type}@{client_id}/{client_secret}/{targets}",
    )

    # Explicit URL tokens we care about (all others from base are ignored)
    template_tokens = dict(NotifyBase.template_tokens, **{
        "type": {
            "name": _("Message Type"),
            "type": "string",
            "regex": (r"^[A-Z0-9_-]+$", "i"),
            "required": True,
            "map_to": "message_type",
        },
        "client_id": {
            "name": _("Client ID"),
            "type": "string",
            "required": True,
        },
        "client_secret": {
            "name": _("Client Secret"),
            "type": "string",
            "required": True,
            "private": True,
        },
        "target_email": {
            "name": _("Target Email"),
            "type": "string",
            "map_to": "targets",
        },
        "target_id": {
            "name": _("Target ID"),
            "type": "string",
            "map_to": "targets",
        },
        "target_sms": {
            "name": _("Target SMS"),
            "type": "string",
            "map_to": "targets",
        },
        "targets": {
            "name": _("Targets"),
            "type": "list:string",
        },
    })

    # Supported query args
    template_args = dict(NotifyBase.template_args, **{
        "type": {
            "alias_of": "type",
        },
        "channels": {
            "name": _("Channels"),
            "type": "list:string",
            "values": NOTIFICATIONAPI_CHANNELS,
        },
        "region": {
            "name": _("Region Name"),
            "type": "choice:string",
            "values": NOTIFICATIONAPI_REGIONS,
            "default": NotificationAPIRegion.US,
        },
        "mode": {
            "name": _("Mode"),
            "type": "choice:string",
            "values": NOTIFICATIONAPI_MODES,
        },
        "to": {
            "alias_of": "targets",
        },
        # Email Values
        "cc": {
            "name": _("Carbon Copy"),
            "type": "list:string",
        },
        "bcc": {
            "name": _("Blind Carbon Copy"),
            "type": "list:string",
        },
        "reply": {
            "name": _("Reply To"),
            "type": "string",
            "map_to": "reply_to",
        },
        "from": {
            "name": _("From Email"),
            "type": "string",
            "map_to": "from_addr",
        },
        "id": {
            "alias_of": "client_id",
        },
        "secret": {
            "alias_of": "client_secret",
        }
    })

    # Define our token control
    template_kwargs = {
        "tokens": {
            "name": _("Template Tokens"),
            "prefix": ":",
        },
    }

    def __init__(self, client_id, client_secret, message_type=None,
                 targets=None, cc=None, bcc=None, reply_to=None,
                 channels=None, region=None, mode=None, from_addr=None,
                 tokens=None, **kwargs):
        """
        Initialize Notify NotificationAPI Object
        """
        super().__init__(**kwargs)

        # Client ID
        self.client_id = validate_regex(client_id)
        if not self.client_id:
            msg = "An invalid NotificationAPI Client ID " \
                  "({}) was specified.".format(client_id)
            self.logger.warning(msg)
            raise TypeError(msg)

        # Client Secret
        self.client_secret = validate_regex(client_secret)
        if not self.client_secret:
            msg = "An invalid NotificationAPI Client Secret " \
                  "({}) was specified.".format(client_secret)
            self.logger.warning(msg)
            raise TypeError(msg)

        # For tracking our email -> name lookups
        self.names = {}

        # Prepare our From Address
        _from_addr = [self.app_id, ""]
        self.from_addr = None
        if isinstance(from_addr, str):
            result = is_email(from_addr)
            if result:
                _from_addr = (
                    result["name"] if result["name"] else _from_addr[0],
                    result["full_email"])
            else:
                # Only update the string but use the already detected info
                _from_addr[0] = from_addr

                # Store our lookup
            self.from_addr = _from_addr[1]
        self.names[_from_addr[1]] = _from_addr[0]

        # Prepare our Reply-To Address
        self.reply_to = {}
        if isinstance(reply_to, str):
            result = is_email(reply_to)
            if result and "full_email" in result:
                self.reply_to = {
                    "senderName": result["name"]
                    if result["name"] else _from_addr[0],
                    "senderEmail": result["full_email"],
                }

        # Our Targets are delimited by found ids
        self.targets = []
        if mode and isinstance(mode, str):
            self.mode = next(
                (a for a in NOTIFICATIONAPI_MODES if a.startswith(mode)), None
            )
            if self.mode not in NOTIFICATIONAPI_MODES:
                msg = \
                    f"The NotificationAPI mode specified ({mode}) is invalid."
                self.logger.warning(msg)
                raise TypeError(msg)
        else:
            # Detect mode based on whether or not a message_type was provided
            self.mode = NotificationAPIMode.MESSAGE if not message_type else \
                NotificationAPIMode.TEMPLATE

        if not message_type:
            # Assign a default message type
            self.message_type = self.default_message_type

        else:
            self.message_type = validate_regex(
                message_type, *self.template_tokens["type"]["regex"])
            if not self.message_type:
                msg = "An invalid NotificationAPI Message Type " \
                      "({}) was specified.".format(message_type)
                self.logger.warning(msg)
                raise TypeError(msg)

        # Precompute auth header
        # Ruby/docs show POST "/{client_id}/sender" with:
        #      Basic base64(client_id:client_secret)
        # https://www.notificationapi.com/docs/reference/server
        token = base64.b64encode(
            f"{self.client_id}:{self.client_secret}".encode()).decode("ascii")
        self.auth_header = f"Basic {token}"

        # Acquire Carbon Copies
        self.cc = set()

        # Acquire Blind Carbon Copies
        self.bcc = set()

        # Store our region
        try:
            self.region = (
                self.template_args["region"]["default"]
                if region is None else region.lower()
            )

            if self.region not in NOTIFICATIONAPI_REGIONS:
                # allow the outer except to handle this common response
                raise IndexError()

        except (AttributeError, IndexError, TypeError):
            # Invalid region specified
            msg = \
                f"The NotificationAPI region specified ({region}) is invalid."
            self.logger.warning(msg)
            raise TypeError(msg) from None

        # Initialize an empty set of channels
        self.channels = set()
        for _channel in parse_list(channels):
            channel = _channel.lower()
            if channel not in NOTIFICATIONAPI_CHANNELS:
                # Invalid channel specified
                msg = (
                    "The NotificationAPI forced channel specified "
                    f"({channel}) is invalid.")
                self.logger.warning(msg)
                raise TypeError(msg) from None
            self.channels.add(channel)

        # Used for URL generation afterwards only
        self._invalid_targets = []

        if targets:
            current_target = {}
            for entry in parse_list(targets, sort=False):
                result = is_email(entry)
                if result:
                    if "email" not in current_target:
                        current_target["email"] = result["full_email"]
                        if not self.channels:
                            self.channels.add(NotificationAPIChannel.EMAIL)
                            self.logger.info(
                                "The NotificationAPI default channel of "
                                f"{NotificationAPIChannel.EMAIL} was set.")
                        continue

                    elif "id" in current_target:
                        # Store and move on
                        self.targets.append(current_target)
                        current_target = {
                            "email": result["full_email"]
                        }
                        continue

                    # if we got here, we have to many emails making it now
                    # ambiguous as to who the sender intended to notify
                    msg = (
                        "The NotificationAPI received too many emails "
                        "creating an ambiguous situation; aborted at "
                        f"'{entry}'.")
                    self.logger.warning(msg)
                    raise TypeError(msg) from None

                result = is_phone_no(entry)
                if result:
                    if "number" not in current_target:
                        current_target["number"] = \
                            ("+" if entry[0] == "+" else "") + result["full"]
                        if not self.channels:
                            self.channels.add(NotificationAPIChannel.SMS)
                            self.logger.info(
                                "The NotificationAPI default channel of "
                                f"{NotificationAPIChannel.SMS} was set.")
                        continue

                    elif "id" in current_target:
                        # Store and move on
                        self.targets.append(current_target)
                        current_target = {
                            "number": result["full"]
                        }
                        continue

                    # if we got here, we have to many emails making it now
                    # ambiguous as to who the sender intended to notify
                    msg = (
                        "The NotificationAPI received too many phone no's "
                        "creating an ambiguous situation; aborted at "
                        f"'{entry}'.")
                    self.logger.warning(msg)
                    raise TypeError(msg) from None

                result = IS_VALID_ID_RE.match(entry)
                if result:
                    if "id" not in current_target:
                        current_target["id"] = result.group("id")
                        continue

                    # Store id in next target and move on
                    self.targets.append(current_target)
                    current_target = {
                        "id": result.group("id")
                    }
                    continue

                self.logger.warning(
                    "Dropped invalid NotificationAPI target "
                    f"({entry}) specified")
                self._invalid_targets.append(entry)
                continue

            if "id" in current_target:
                # Store our final entry
                self.targets.append(current_target)
                current_target = {}

            if current_target:
                # we have email or sms, but no id to go with it
                msg = (
                    "The NotificationAPI did not detect an id to "
                    "correlate the following with {}".format(
                        str(current_target)))
                self.logger.warning(msg)
                raise TypeError(msg) from None

        # Validate recipients (cc:) and drop bad ones:
        for recipient in parse_emails(cc):
            result = is_email(recipient)
            if result:
                self.cc.add(result["full_email"])
                if result["name"]:
                    self.names[result["full_email"]] = result["name"]
                continue

            self.logger.warning(
                "Dropped invalid Carbon Copy email "
                "({}) specified.".format(recipient),
            )

        # Validate recipients (bcc:) and drop bad ones:
        for recipient in parse_emails(bcc):

            result = is_email(recipient)
            if result:
                self.bcc.add(result["full_email"])
                if result["name"]:
                    self.names[result["full_email"]] = result["name"]
                continue

            self.logger.warning(
                "Dropped invalid Blind Carbon Copy email "
                "({}) specified.".format(recipient),
            )

        # Template functionality
        self.tokens = {}
        if isinstance(tokens, dict):
            self.tokens.update(tokens)

        return

    @property
    def url_identifier(self):
        """
        Returns all of the identifiers that make this URL unique from
        another similar one. Targets or end points should never be identified
        here.
        """
        return (self.secure_protocol[0], self.client_id, self.client_secret)

    def url(self, privacy=False, *args, **kwargs):
        """
        Returns the URL built dynamically based on specified arguments.
        """

        # Define any URL parameters
        params = {
            "mode": self.mode,
        }

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        if len(self.cc) > 0:
            # Handle our Carbon Copy Addresses
            params["cc"] = ",".join([
                formataddr(
                    (self.names.get(e, False), e),
                    # Swap comma for it's escaped url code (if detected) since
                    # we're using that as a delimiter
                    charset="utf-8").replace(",", "%2C")
                for e in self.cc])

        if len(self.bcc) > 0:
            # Handle our Blind Carbon Copy Addresses
            params["bcc"] = ",".join([
                formataddr(
                    (self.names.get(e, False), e),
                    # Swap comma for it's escaped url code (if detected) since
                    # we're using that as a delimiter
                    charset="utf-8").replace(",", "%2C")
                for e in self.bcc])

        if self.reply_to:
            # Handle our Reply-To Address
            params["reply"] = formataddr(
                (self.reply_to["senderName"], self.reply_to["senderEmail"]),
                # Swap comma for its escaped url code (if detected) since
                # we're using that as a delimiter
                charset="utf-8",
            )

        if self.channels:
            # Prepare our default channel
            params["channels"] = ",".join(self.channels)

        if self.region != self.template_args["region"]["default"]:
            # Prepare our default region
            params["region"] = self.region

        # handle from=
        if self.from_addr and self.names[self.from_addr] != self.app_id:
            params["from"] = self.names[self.from_addr]

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))
        # Store any template entries if specified
        params.update({f":{k}": v for k, v in self.tokens.items()})

        targets = []
        for target in self.targets:
            # ID is always present
            targets.append(f"@{target['id']}")
            if "number" in target:
                targets.append(f"{target['number']}")
            if "email" in target:
                targets.append(f"{target['email']}")

        mtype = f"{self.message_type}@" \
            if self.message_type != self.default_message_type else ""
        return "{schema}://{mtype}{cid}/{secret}/{targets}?{params}".format(
            schema=self.secure_protocol[0],
            mtype=mtype,
            cid=self.pprint(self.client_id, privacy, safe=""),
            secret=self.pprint(self.client_secret, privacy, safe=""),
            targets=NotifyNotificationAPI.quote("/".join(
                chain(targets, self._invalid_targets)), safe="/"),
            params=NotifyNotificationAPI.urlencode(params),
        )

    def __len__(self):
        """
        Returns the number of targets associated with this notification
        """

        return max(1, len(self.targets))

    def gen_payload(self, body, title="", notify_type=NotifyType.INFO,
                    **kwargs):
        """
        generates our NotificationAPI payload
        """

        _payload = {
            "type": self.message_type,
        }
        if self.mode == NotificationAPIMode.TEMPLATE:
            # Take a copy of our token dictionary
            parameters = self.tokens.copy()

            # Apply some defaults template values
            parameters["appBody"] = body
            parameters["appTitle"] = title
            parameters["appType"] = notify_type.value
            parameters["appId"] = self.app_id
            parameters["appDescription"] = self.app_desc
            parameters["appColor"] = self.color(notify_type)
            parameters["appImageUrl"] = self.image_url(notify_type)
            parameters["appUrl"] = self.app_url

            # A Simple Email Payload Template
            _payload.update({
                "parameters": {**parameters},
            })

        else:
            # Acquire text version of body if provided
            text_body = convert_between(
                NotifyFormat.HTML, NotifyFormat.TEXT, body) \
                if self.notify_format == NotifyFormat.HTML else body

            for channel in self.channels:
                # Python v3.10 supports `match/case` but since Apprise aims to
                # be compatible with Python v3.9+, we must use if/else for the
                # time being
                if channel == NotificationAPIChannel.SMS:
                    _payload.update({
                        NotificationAPIChannel.SMS: {
                            "message": (title + "\n" + text_body)
                            if title else text_body,
                        },
                    })

                elif channel == NotificationAPIChannel.EMAIL:
                    html_body = convert_between(
                        NotifyFormat.TEXT, NotifyFormat.HTML, body) \
                        if self.notify_format != NotifyFormat.HTML else body

                    _payload.update({
                        NotificationAPIChannel.EMAIL: {
                            "subject": title if title else self.app_id,
                            "html": html_body,
                        },
                    })

                    if self.from_addr:
                        _payload[NotificationAPIChannel.EMAIL].update({
                            "senderEmail": self.from_addr,
                            "senderName": self.names[self.from_addr],
                        })

                elif channel == NotificationAPIChannel.INAPP:
                    _payload.update({
                        NotificationAPIChannel.INAPP: {
                            "title": title if title else self.app_id,
                            "image": self.image_url(notify_type),
                        },
                    })

                elif channel == NotificationAPIChannel.WEB_PUSH:
                    _payload.update({
                        NotificationAPIChannel.WEB_PUSH: {
                            "title": title if title else self.app_id,
                            "message": text_body,
                            "icon": self.image_url(notify_type),
                        },
                    })

                elif channel == NotificationAPIChannel.MOBILE_PUSH:
                    _payload.update({
                        NotificationAPIChannel.MOBILE_PUSH: {
                            "title": title if title else self.app_id,
                            "message": text_body,
                        },
                    })

                else:  # channel == NotificationAPIChannel.SLACK
                    _payload.update({
                        NotificationAPIChannel.SLACK: {
                            "text": (title + "\n" + text_body)
                            if title else text_body,
                        },
                    })

        # Copy our list to work with
        targets = list(self.targets)
        if self.from_addr:
            _payload.update({
                "options": {
                    "email": {
                        "fromAddress": self.from_addr,
                        "fromName": self.names[self.from_addr]}}})

        elif self.cc or self.bcc:
            # Set up shell
            _payload.update({"options": {"email": {}}})

        while len(targets) > 0:
            target = targets.pop(0)

            # Create a copy of our template
            payload = _payload.copy()

            # the cc, bcc, to field must be unique or SendMail will fail,
            # the below code prepares this by ensuring the target isn't in
            # the cc list or bcc list. It also makes sure the cc list does
            # not contain any of the bcc entries
            if "email" in target:
                cc = (self.cc - self.bcc - {target["email"]})
                bcc = (self.bcc - {target["email"]})

            else:
                # Assume defaults
                cc = self.cc
                bcc = self.bcc

            #
            # Prepare our 'to'
            #
            payload["to"] = {**target}

            # Support cc/bcc
            if len(cc):
                payload["options"]["email"]["ccAddresses"] = list(cc)
            if len(bcc):
                payload["options"]["email"]["bccAddresses"] = list(bcc)

            yield payload

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """
        Perform NotificationAPI Notification
        """

        # error tracking (used for function return)
        has_error = False

        if not self.targets:
            # There is no one to email or send an sms message to; we're done
            self.logger.warning(
                "There are no NotificationAPI recipients to notify"
            )
            return False

        # Prepare our URL
        url = (
            f"{NOTIFICATIONAPI_API_LOOKUP[self.region]}/"
            f"{self.client_id}/sender")

        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json",
            "Authorization": self.auth_header,
        }

        for payload in self.gen_payload(
                body, title=title, notify_type=notify_type, **kwargs):
            # Perform our post
            self.logger.debug(
                "NotificationAPI POST URL: {} (cert_verify={!r})".format(
                    url, self.verify_certificate))
            self.logger.debug(
                "NotificationAPI Payload: %s", str(payload["to"]["id"]))

            # Always call throttle before any remote server i/o is made
            self.throttle()
            try:
                r = requests.post(
                    url,
                    data=dumps(payload),
                    headers=headers,
                    verify=self.verify_certificate,
                    timeout=self.request_timeout,
                )

                try:
                    loads(r.content)

                except (AttributeError, TypeError, ValueError):
                    # This gets thrown if we can't parse our JSON Response
                    #  - ValueError = r.content is Unparsable
                    #  - TypeError = r.content is None
                    #  - AttributeError = r is None
                    self.logger.warning(
                        "Invalid response from NotificationAPI server.")
                    self.logger.debug(
                        "Response Details:\r\n{}".format(r.content))

                    # Record our failure
                    has_error = True
                    continue

                # Reference status code
                status_code = r.status_code

                if status_code not in (
                        requests.codes.ok, requests.codes.accepted):
                    # We had a problem
                    status_str = \
                        NotifyNotificationAPI.http_response_code_lookup(
                            status_code)

                    self.logger.warning(
                        "Failed to send NotificationAPI notification to %s: "
                        "%s%serror=%d",
                        payload["to"]["id"],
                        status_str,
                        ", " if status_str else "",
                        status_code)
                    self.logger.debug(
                        "Response Details:\r\n%s", str(r.content))

                    # Record our failure
                    has_error = True

                else:
                    self.logger.info(
                        "Sent NotificationAPI notification to %s.",
                        payload["to"]["id"])

            except requests.RequestException as e:
                self.logger.warning(
                    "A Connection error occurred sending NotificationAPI "
                    "notification to %s.", payload["to"]["id"])
                self.logger.debug("Socket Exception: {}".format(str(e)))

                # Record our failure
                has_error = True

        return not has_error

    @staticmethod
    def parse_url(url):
        """
        Parses the URL and returns enough arguments that can allow
        us to re-instantiate this object.

        """

        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Define our minimum requirements; defining them now saves us from
        # having to if/else all kinds of branches below...
        results["client_id"] = None
        results["client_secret"] = None

        # Prepare our targets (starting with our host)
        results["targets"] = []
        if results["host"]:
            results["targets"].append(
                NotifyNotificationAPI.unquote(results["host"]))

        # For tracking email sources
        results["from_addr"] = None
        if "from" in results["qsd"] and len(results["qsd"]["from"]):
            results["from_addr"] = \
                NotifyNotificationAPI.unquote(results["qsd"]["from"].rstrip())

        # First 2 elements are the client_id and client_secret
        # Following are targets
        results["targets"] += \
            NotifyNotificationAPI.split_path(results["fullpath"])
        # check for our client id
        if "id" in results["qsd"] and len(results["qsd"]["id"]):
            # Store our Client ID
            results["client_id"] = \
                NotifyNotificationAPI.unquote(results["qsd"]["id"])

        elif results["targets"]:
            # Store our Client ID
            results["client_id"] = results["targets"].pop(0)

        if "secret" in results["qsd"] and len(results["qsd"]["secret"]):
            # Store our Client Secret
            results["client_secret"] = \
                NotifyNotificationAPI.unquote(results["qsd"]["secret"])

        elif results["targets"]:
            # Store our Client Secret
            results["client_secret"] = results["targets"].pop(0)

        if "region" in results["qsd"] and len(results["qsd"]["region"]):
            results["region"] = \
                NotifyNotificationAPI.unquote(results["qsd"]["region"])

        if "channels" in results["qsd"] and len(results["qsd"]["channels"]):
            results["channels"] = \
                NotifyNotificationAPI.unquote(results["qsd"]["channels"])

        if "mode" in results["qsd"] and len(results["qsd"]["mode"]):
            results["mode"] = \
                NotifyNotificationAPI.unquote(results["qsd"]["mode"])

        if "reply" in results["qsd"] and len(results["qsd"]["reply"]):
            results["reply_to"] = \
                NotifyNotificationAPI.unquote(results["qsd"]["reply"])

        # Handling of Message Type
        if "type" in results["qsd"] and len(results["qsd"]["type"]):
            results["message_type"] = \
                NotifyNotificationAPI.unquote(results["qsd"]["type"])

        elif results["user"]:
            # Pull from user
            results["message_type"] = \
                NotifyNotificationAPI.unquote(results["user"])

        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"].append(
                NotifyNotificationAPI.unquote(results["qsd"]["to"]))

        # Handle Carbon Copy Addresses
        if "cc" in results["qsd"] and len(results["qsd"]["cc"]):
            results["cc"] = \
                NotifyNotificationAPI.unquote(results["qsd"]["cc"])

        # Handle Blind Carbon Copy Addresses
        if "bcc" in results["qsd"] and len(results["qsd"]["bcc"]):
            results["bcc"] = \
                NotifyNotificationAPI.unquote(results["qsd"]["bcc"])

        # Store our tokens
        results["tokens"] = results["qsd:"]

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# API Reference: https://developers.brevo.com/reference/getting-started-1

from json import dumps
from os.path import splitext

import requests

from .. import exception
from ..common import NotifyFormat, NotifyType
from ..conversion import convert_between
from ..locale import gettext_lazy as _
from ..utils.parse import is_email, parse_list, validate_regex
from .base import NotifyBase

# Extend HTTP Error Messages (most common Brevo SMTP errors)
BREVO_HTTP_ERROR_MAP = {
    400: "Bad Request - Invalid payload or missing parameters.",
    401: "Unauthorized - Invalid Brevo API key.",
    402: "Payment Required - Plan limitation or credit issue.",
    429: "Too Many Requests - Rate limit exceeded.",
}

# Comprehensive list of Brevo-supported extensions for Transactional Emails
# Source: Brevo API Documentation & Transactional Attachment Guidelines
BREVO_VALID_EXTENSIONS = (
    # Documents & Text
    "xlsx", "xls", "ods", "docx", "docm", "doc", "csv", "pdf", "txt",
    "rtf", "msg", "pub", "mobi", "ppt", "pptx", "eps", "odt", "ics",
    "xml", "css", "html", "htm", "shtml",
    # Images
    "gif", "jpg", "jpeg", "png", "tif", "tiff", "bmp", "cgm",
    # Archives
    "zip", "tar", "ez", "pkpass",
    # Audio
    "mp3", "m4a", "m4v", "wma", "ogg", "flac", "wav", "aif", "aifc", "aiff",
    # Video
    "mp4", "mov", "avi", "mkv", "mpeg", "mpg", "wmv"
)


class NotifyBrevo(NotifyBase):
    """A wrapper for Notify Brevo Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Brevo"

    # The services URL
    service_url = "https://www.brevo.com/"

    # The default secure protocol
    secure_protocol = "brevo"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_brevo"

    # Default to markdown
    notify_format = NotifyFormat.HTML

    # The default Email API URL to use
    notify_url = "https://api.brevo.com/v3/smtp/email"

    # Support attachments
    attachment_support = True

    # Allow 300 requests per minute.
    # 60/300 = 0.2
    request_rate_per_sec = 0.2

    # The default subject to use if one isn't specified.
    default_empty_subject = "<no subject>"

    # Define object templates
    templates = (
        "{schema}://{apikey}:{from_email}",
        "{schema}://{apikey}:{from_email}/{targets}",
    )

    # Define our template arguments
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "apikey": {
                "name": _("API Key"),
                "type": "string",
                "private": True,
                "required": True,
                "regex": (r"^[a-zA-Z0-9._-]+$", "i"),
            },
            "from_email": {
                "name": _("Source Email"),
                "type": "string",
                "required": True,
            },
            "target_email": {
                "name": _("Target Email"),
                "type": "string",
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "to": {
                "alias_of": "targets",
            },
            "cc": {
                "name": _("Carbon Copy"),
                "type": "list:string",
            },
            "bcc": {
                "name": _("Blind Carbon Copy"),
                "type": "list:string",
            },
            "reply": {
                "name": _("Reply To Email"),
                "type": "string",
                "map_to": "reply_to",
            },
        },
    )

    def __init__(
        self,
        apikey,
        from_email,
        targets=None,
        reply_to=None,
        cc=None,
        bcc=None,
        **kwargs,
    ):
        """Initialize Notify Brevo Object."""
        super().__init__(**kwargs)

        # API Key (associated with project)
        self.apikey = validate_regex(
            apikey, *self.template_tokens["apikey"]["regex"]
        )
        if not self.apikey:
            msg = f"An invalid Brevo API Key ({apikey}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        result = is_email(from_email)
        if not result:
            msg = f"Invalid ~From~ email specified: {from_email}"
            self.logger.warning(msg)
            raise TypeError(msg)

        # Store email address
        self.from_email = result["full_email"]

        # Reply-to
        self.reply_to = None
        if reply_to:
            result = is_email(reply_to)
            if not result:
                msg = "An invalid Brevo Reply To ({}) was specified.".format(
                    f"{reply_to}")
                self.logger.warning(msg)
                raise TypeError(msg)

            self.reply_to = (
                    result["name"] if result["name"] else False,
                    result["full_email"],
                )

        # Acquire Targets (To Emails)
        self.targets = []

        # Acquire Carbon Copies
        self.cc = set()

        # Acquire Blind Carbon Copies
        self.bcc = set()

        # Validate recipients (to:) and drop bad ones:
        if targets:
            for recipient in parse_list(targets):

                result = is_email(recipient)
                if result:
                    self.targets.append(result["full_email"])
                    continue

                self.logger.warning(
                    f"Dropped invalid email ({recipient}) specified.",
                )
        else:
            # add ourselves
            self.targets.append(self.from_email)

        # Validate recipients (cc:) and drop bad ones:
        for recipient in parse_list(cc):

            result = is_email(recipient)
            if result:
                self.cc.add(result["full_email"])
                continue

            self.logger.warning(
                f"Dropped invalid Carbon Copy email ({recipient}) specified.",
            )

        # Validate recipients (bcc:) and drop bad ones:
        for recipient in parse_list(bcc):

            result = is_email(recipient)
            if result:
                self.bcc.add(result["full_email"])
                continue

            self.logger.warning(
                "Dropped invalid Blind Carbon Copy email "
                f"({recipient}) specified.",
            )

        return

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (self.secure_protocol, self.apikey, self.from_email)

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Our URL parameters
        params = self.url_parameters(privacy=privacy, *args, **kwargs)

        if len(self.cc) > 0:
            # Handle our Carbon Copy Addresses
            params["cc"] = ",".join(self.cc)

        if len(self.bcc) > 0:
            # Handle our Blind Carbon Copy Addresses
            params["bcc"] = ",".join(self.bcc)

        if self.reply_to:
            # Handle our reply to address
            params["reply"] = (
                "{} <{}>".format(*self.reply_to)
                if self.reply_to[0]
                else self.reply_to[1]
            )

        # a simple boolean check as to whether we display our target emails
        # or not
        has_targets = not (
            len(self.targets) == 1 and self.targets[0] == self.from_email
        )

        return "{schema}://{apikey}:{from_email}/{targets}?{params}".format(
            schema=self.secure_protocol,
            apikey=self.pprint(self.apikey, privacy, safe=""),
            # never encode email since it plays a huge role in our hostname
            from_email=self.from_email,
            targets=(
                ""
                if not has_targets
                else "/".join(
                    [NotifyBrevo.quote(x, safe="") for x in self.targets]
                )
            ),
            params=NotifyBrevo.urlencode(params),
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        return max(len(self.targets), 1)

    def send(
        self,
        body,
        title="",
        notify_type=NotifyType.INFO,
        attach=None,
        **kwargs,
    ):
        """Perform Brevo Notification."""

        if not self.targets:
            # There is no one to email; we're done
            self.logger.warning(
                "There are no Brevo email recipients to notify")
            return False

        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json",
            "Accept": "application/json",
            "api-key": self.apikey,
        }

        # error tracking (used for function return)
        has_error = False

        # A Simple Email Payload Template
        _payload = {
            "sender": {
                "email": self.from_email,
            },
            # Placeholder, filled per target
            "to": [{"email": None}],
            "subject": title if title else self.default_empty_subject,
        }
        # Body selection
        use_html = self.notify_format == NotifyFormat.HTML

        if use_html:
            # body already normalised; keep your existing logic
            _payload["htmlContent"] = body
            _payload["textContent"] = convert_between(
                NotifyFormat.HTML, NotifyFormat.TEXT, body
            )
        else:
            # Plain text requested, but Brevo still wants HTML
            _payload["textContent"] = body
            _payload["htmlContent"] = convert_between(
                NotifyFormat.TEXT, NotifyFormat.HTML, body
            )

        if attach and self.attachment_support:
            attachments = []

            # Send our attachments
            for no, attachment in enumerate(attach, start=1):
                # Perform some simple error checking
                if not attachment:
                    # We could not access the attachment
                    self.logger.error(
                        "Could not access Brevo attachment"
                        f" {attachment.url(privacy=True)}."
                    )
                    return False

                # Brevo does not track content/mime type and relies 100%
                # entirely on the filename extension as to whether or not it
                # will accept it or not.
                #
                # The below prepares a safe_name (which can't be .dat like
                # other plugins since Brevo rejects that type). For this
                # reason .txt is chosen intentionally for this circumstance.

                # Use the attachment name if available, otherwise default to a
                # generic name
                raw_name = attachment.name \
                    if attachment.name else f"file{no:03}.txt"

                # If the filename does NOT match a supported extension, append
                # .txt
                _, ext = splitext(raw_name)
                safe_name = f"{raw_name}.txt" if (
                    not ext or ext[1:].lower()
                    not in BREVO_VALID_EXTENSIONS) else raw_name

                try:
                    attachments.append({
                        "content": attachment.base64(),
                        "name": safe_name,
                    })

                except exception.AppriseException:
                    # We could not access the attachment
                    self.logger.error(
                        "Could not access Brevo attachment"
                        f" {attachment.url(privacy=True)}."
                    )
                    return False

                self.logger.debug(
                    "Appending Brevo attachment"
                    f" {attachment.url(privacy=True)}"
                )

            # Append our attachments to the payload
            _payload.update({
                "attachment": attachments,
            })

        if self.reply_to:
            _payload["replyTo"] = {"email": self.reply_to[1]}

        targets = list(self.targets)
        while len(targets) > 0:
            target = targets.pop(0)

            # Create a copy of our template
            payload = _payload.copy()

            # the cc, bcc, to field must be unique or SendMail will fail, the
            # below code prepares this by ensuring the target isn't in the cc
            # list or bcc list. It also makes sure the cc list does not contain
            # any of the bcc entries
            cc = self.cc - self.bcc - {target}
            bcc = self.bcc - {target}

            # Set our main recipient
            payload["to"] = [{"email": target}]

            if len(cc):
                payload["cc"] = [{"email": email} for email in cc]

            if len(bcc):
                payload["bcc"] = [{"email": email} for email in bcc]

            self.logger.debug(
                "Brevo POST URL:"
                f" {self.notify_url} (cert_verify={self.verify_certificate!r})"
            )
            self.logger.debug(f"Brevo Payload: {payload!s}")

            # Always call throttle before any remote server i/o is made
            self.throttle()
            try:
                r = requests.post(
                    self.notify_url,
                    data=dumps(payload),
                    headers=headers,
                    verify=self.verify_certificate,
                    timeout=self.request_timeout,
                )
                if r.status_code not in (
                    requests.codes.ok,
                    requests.codes.accepted,
                    requests.codes.created,
                ):
                    # We had a problem
                    status_str = NotifyBrevo.http_response_code_lookup(
                        r.status_code, BREVO_HTTP_ERROR_MAP
                    )

                    self.logger.warning(
                        "Failed to send Brevo notification to {}: "
                        "{}{}error={}.".format(
                            target,
                            status_str,
                            ", " if status_str else "",
                            r.status_code,
                        )
                    )

                    self.logger.debug(f"Response Details:\r\n{r.content}")

                    # Mark our failure
                    has_error = True
                    continue

                else:
                    self.logger.info(
                        f"Sent Brevo notification to {target}."
                    )

            except requests.RequestException as e:
                self.logger.warning(
                    "A Connection error occurred sending Brevo "
                    f"notification to {target}."
                )
                self.logger.debug(f"Socket Exception: {e!s}")

                # Mark our failure
                has_error = True
                continue

        return not has_error

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""

        results = NotifyBase.parse_url(url)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Our URL looks like this:
        #    {schema}://{apikey}:{from_email}/{targets}
        #
        # which actually equates to:
        #    {schema}://{user}:{password}@{host}/{email1}/{email2}/etc..
        #                 ^       ^         ^
        #                 |       |         |
        #              apikey     -from addr-

        if not results.get("user"):
            # An API Key as not properly specified
            return None

        if not results.get("password"):
            # A From Email was not correctly specified
            return None

        # Prepare our API Key
        results["apikey"] = NotifyBrevo.unquote(results["user"])

        # Prepare our From Email Address
        results["from_email"] = "{}@{}".format(
            NotifyBrevo.unquote(results["password"]),
            NotifyBrevo.unquote(results["host"]),
        )

        # Acquire our targets
        results["targets"] = NotifyBrevo.split_path(results["fullpath"])

        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += NotifyBrevo.parse_list(
                results["qsd"]["to"]
            )

        # Handle Carbon Copy Addresses
        if "cc" in results["qsd"] and len(results["qsd"]["cc"]):
            results["cc"] = NotifyBrevo.parse_list(results["qsd"]["cc"])

        # Handle Blind Carbon Copy Addresses
        if "bcc" in results["qsd"] and len(results["qsd"]["bcc"]):
            results["bcc"] = NotifyBrevo.parse_list(results["qsd"]["bcc"])

        # Handle Reply To Address
        if "reply" in results["qsd"] and len(results["qsd"]["reply"]):
            results["reply_to"] = NotifyBrevo.unquote(results["qsd"]["reply"])

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# API Refererence:
#   - https://docs.signl4.com/integrations/webhook/webhook.html
#

from json import dumps
from typing import Any, Optional

import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from ..utils.parse import parse_bool, validate_regex
from .base import NotifyBase


class NotifySIGNL4(NotifyBase):
    """
    A wrapper for SIGNL4 Notifications
    """

    # The default descriptive name associated with the Notification
    service_name = "SIGNL4"

    # The services URL
    service_url = "https://signl4.com/"

    # Secure Protocol
    secure_protocol = "signl4"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_signl4"

    # Our event action type
    event_action = "trigger"

    # Our default notification URL
    notify_url = "https://connect.signl4.com/webhook/{secret}/"

    # Define object templates
    templates = (
        "{schema}://{secret}",
    )

    # Define our template tokens
    template_tokens = dict(NotifyBase.template_tokens, **{
        # SIGNL4 team or integration secret
        "secret": {
            "name": _("Secret"),
            "type": "string",
            "private": True,
            "required": True
        },
    })

    # Define our template arguments
    template_args = dict(NotifyBase.template_args, **{
        "service": {
            "name": _("Service"),
            "type": "string",
        },
        "location": {
            "name": _("Location"),
            "type": "string",
        },
        "alerting_scenario": {
            "name": _("Alerting Scenario"),
            "type": "string",
        },
        "filtering": {
            "name": _("Filtering"),
            "type": "bool",
            "default": False,
        },
        "external_id": {
            "name": _("External ID"),
            "type": "string",
        },
        "status": {
            "name": _("Status"),
            "type": "string",
        },
    })

    def __init__(
        self,
        secret: str,
        service: Optional[str] = None,
        location: Optional[str] = None,
        alerting_scenario: Optional[str] = None,
        filtering: Optional[bool] = None,
        external_id: Optional[str] = None,
        status: Optional[str] = None,
        **kwargs: Any
    ) -> None:
        """
        Initialize SIGNL4 Object
        """
        super().__init__(**kwargs)

        # SIGNL4 team or integration secret
        self.secret = validate_regex(secret)
        if not self.secret:
            msg = "An invalid SIGNL4 team or integration secret " \
                  "({}) was specified.".format(secret)
            self.logger.warning(msg)
            raise TypeError(msg)

        # A service option for notifications
        self.service = service

        # A location option for notifications
        self.location = location

        # A alerting_scenario option for notifications
        self.alerting_scenario = alerting_scenario

        # A filtering option for notifications
        self.filtering = (
            self.template_args["filtering"]["default"]
            if filtering is None
            else bool(filtering)
        )

        # A external_id option for notifications
        self.external_id = external_id

        # A location option for notifications
        self.status = status

        return

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """
        Send our SIGNL4 Notification
        """

        # Prepare our headers
        headers = {
            "Content-Type": "application/json",
        }

        # Prepare our persistent_notification.create payload
        payload = {
            "title": title if title else self.app_desc,
            "body": body,
            "X-S4-SourceSystem": self.app_id,
        }

        if self.service:
            payload["X-S4-Service"] = self.service

        if self.alerting_scenario:
            payload["X-S4-AlertingScenario"] = self.alerting_scenario

        if self.location:
            payload["X-S4-Location"] = self.location

        if self.filtering:
            payload["X-S4-Filtering"] = self.filtering

        if self.external_id:
            payload["X-S4-ExternalID"] = self.external_id

        if self.status:
            payload["X-S4-Status"] = self.status

        # Prepare our URL
        notify_url = self.notify_url.format(secret=self.secret)

        self.logger.debug(
            "SIGNL4 POST URL: %s (cert_verify=%s)",
            notify_url, self.verify_certificate)
        self.logger.debug("SIGNL4 Payload: %r", payload)


        # Always call throttle before any remote server i/o is made
        self.throttle()

        try:
            r = requests.post(
                notify_url,
                data=dumps(payload),
                headers=headers,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )
            if r.status_code not in (
                    requests.codes.ok, requests.codes.created,
                    requests.codes.accepted):
                # We had a problem
                status_str = \
                    NotifySIGNL4.http_response_code_lookup(
                        r.status_code)

                self.logger.warning(
                    "Failed to send SIGNL4 notification: "
                    "{}{}error={}.".format(
                        status_str,
                        ", " if status_str else "",
                        r.status_code))

                self.logger.debug("Response Details:\r\n%r", r.content)

                # Return; we're done
                return False

            else:
                self.logger.info("Sent SIGNL4 notification.")

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred sending SIGNL4 "
                "notification to %s", self.host)
            self.logger.debug("Socket Exception: %s", str(e))

            # Return; we're done
            return False

        return True

    @property
    def url_identifier(self):
        """
        Returns all of the identifiers that make this URL unique from
        another simliar one. Targets or end points should never be identified
        here.
        """
        return (
            self.secure_protocol, self.secret,
        )

    def url(self, privacy=False, *args, **kwargs):
        """
        Returns the URL built dynamically based on specified arguments.
        """

        # Define any URL parameters
        params = {}

        if self.service is not None:
            params["service"] = self.service

        if self.location is not None:
            params["location"] = self.location

        if self.alerting_scenario is not None:
            params["alerting_scenario"] = self.alerting_scenario

        if self.filtering != self.template_args["filtering"]["default"]:
            # Only add filtering if it is not the default value
            params["filtering"] = "yes" if self.filtering else "no"

        if self.external_id is not None:
            params["external_id"] = self.external_id

        if self.status is not None:
            params["status"] = self.status

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        url = "{schema}://{secret}"

        return url.format(
            schema=self.secure_protocol,
            # never encode hostname since we're expecting it to be a valid one
            secret=self.pprint(
                self.secret, privacy, mode=PrivacyMode.Secret, safe=""),
            params=NotifySIGNL4.urlencode(params),
        )

    @staticmethod
    def parse_url(url):
        """
        Parses the URL and returns enough arguments that can allow
        us to re-instantiate this object.
        """

        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn"t load the results
            return results

        # The "secret" makes it easier to use yaml configuration
        if "secret" in results["qsd"] and \
                len(results["qsd"]["secret"]):
            results["secret"] = \
                NotifySIGNL4.unquote(results["qsd"]["secret"])
        else:
            results["secret"] = \
                NotifySIGNL4.unquote(results["host"])

        if "service" in results["qsd"] and len(results["qsd"]["service"]):
            results["service"] = \
                NotifySIGNL4.unquote(results["qsd"]["service"])

        if "location" in results["qsd"] and len(results["qsd"]["location"]):
            results["location"] = \
                NotifySIGNL4.unquote(results["qsd"]["location"])

        if "alerting_scenario" in results["qsd"] and \
            len(results["qsd"]["alerting_scenario"]):
            results["alerting_scenario"] = \
                NotifySIGNL4.unquote(results["qsd"]["alerting_scenario"])

        if "filtering" in results["qsd"] and len(results["qsd"]["filtering"]):
            results["filtering"] = \
                parse_bool(
                    NotifySIGNL4.unquote(
                        results["qsd"]["filtering"],
                        NotifySIGNL4.template_args["filtering"]["default"]))

        if "external_id" in results["qsd"] and \
            len(results["qsd"]["external_id"]):
            results["external_id"] = \
                NotifySIGNL4.unquote(results["qsd"]["external_id"])

        if "status" in results["qsd"] and len(results["qsd"]["status"]):
            results["status"] = \
                NotifySIGNL4.unquote(results["qsd"]["status"])

        return results
